#include "hmbdc/app/Base.hpp"
#include "hmbdc/pattern/MonoLockFreeBuffer.hpp"
#include "hmbdc/time/Time.hpp"
#include "hmbdc/time/Rater.hpp"
#include "hmbdc/os/Signals.hpp"
#include "hmbdc/numeric/StatHistogram.hpp"

#include <boost/program_options.hpp>
#include <boost/range/combine.hpp>
#include <iostream>
#include <functional>
#include <memory>
#include <unistd.h>
#include <sys/mman.h>
 #include <signal.h>

using namespace hmbdc;
using namespace hmbdc::app;
using namespace hmbdc::time;
using namespace hmbdc::numeric;
using namespace std;

struct Message
: hasTag<1001> {
    Message() 
    : sendTs(SysTime::now())
    {}
    time::SysTime sendTs;
};

namespace {
    bool part;
    uint16_t startCore;
    size_t skipFirst;
    int senderCount;
    int receiverCount;
    uint16_t bufferSizePower2;
    size_t msgPerSec;
    char ipcRole;
}

template <typename Context>
struct SenderClient 
: Client<SenderClient<Context>>
, TimerManager
, ReoccuringTimer {
    using ptr = std::shared_ptr<SenderClient>;
    SenderClient(Context& ctx, uint16_t id, size_t msgPerSec)
    : ReoccuringTimer(Duration::seconds(1u))
    , ctx_(ctx)
    , id_(id)
    , periodicMessageCount_(0)
    , rater_(Duration::seconds(1u), msgPerSec, 1, msgPerSec != 0) {
        rawMessageSize_ = sizeof(MessageWrap<Message>);
        setCallback(
            [this](TimerManager& tm, SysTime const& now) {
            report();
            }
        );
        schedule(SysTime::now(), *this);
    }

    char const* hmbdcName() const { 
        return "perf-tx"; 
    }

    /*virtual*/ 
    void invokedCb(uint16_t) HMBDC_RESTRICT override {
        if (rater_.check()) {
            ctx_.template sendInPlace<Message>();
            rater_.commit();
            periodicMessageCount_++;
        }
    }

    void stoppedCb(std::exception const& e) override {
        std::cout << (e.what()) << std::endl;
    };

    void report() {
        HMBDC_LOG_n("sender ", id_, ":msgSize=", sizeof(Message), " rawMessageSize="
            , rawMessageSize_, " mps=", periodicMessageCount_);
        periodicMessageCount_ = 0u;
    }

    Context& ctx_;
    uint16_t id_;
    size_t periodicMessageCount_;
    size_t rawMessageSize_;
    Rater rater_;
};


using Hist = StatHistogram<Duration, false>;
template <typename Context>
struct ReceiverClient 
: Client<ReceiverClient<Context>, Message>
, TimerManager
, ReoccuringTimer {
    using ptr = std::shared_ptr<ReceiverClient>;

    ReceiverClient(Context& ctx, uint16_t id)
    : ReoccuringTimer(Duration::seconds(1u))
    , ctx_(ctx)
    , id_(id)
    , periodicMessageCount_(0)
    , hist_(Duration::microseconds(0), Duration::microseconds(1000), 10000) {
        setCallback(
            [this](TimerManager& tm, SysTime const& now) {
            report();
            }
        );
        schedule(SysTime::now(), *this);
    }

    char const* hmbdcName() const { 
        return "perf-rx"; 
    }

    void handleMessageCb(Message const& r) {
        periodicMessageCount_++;
        if (!skipFirst) {
            hist_.add(SysTime::now() - r.sendTs);
        } else {
            skipFirst--;
        }
    }

    void stoppedCb(std::exception const& e) override {
        std::cout << (e.what()) << std::endl;
    };

    void report() {
        HMBDC_LOG_n("receiver ", id_, ": mps=", periodicMessageCount_
        );

        periodicMessageCount_ = 0u;
    }

    void showHist() {
        cout << "\nlatencies receiver " << id_ << ':';
        hist_.display(cout, {0, 1, 10, 50, 75, 90, 99, 99.9, 99.99, 100});
        cout << endl;
    }

    Context& ctx_;
    uint16_t id_;
    size_t periodicMessageCount_;
    Hist hist_;
};


template <typename Ctx>
void run() {    
    SingletonGuardian<HMBDC_LOGGER> logGuard(std::cout);
    Ctx ctx(bufferSizePower2);

    std::vector<typename ReceiverClient<Ctx>::ptr> receivers;
    uint64_t cpuMask = 1ul << startCore;

    while (receiverCount-- > 0) {
        auto r = new ReceiverClient<Ctx>(ctx, receiverCount);
        receivers.emplace_back(r);
        ctx.start(*r, cpuMask, true);
        cpuMask <<= 1ul;
    }
    std::vector<typename SenderClient<Ctx>::ptr> senders;
    while (senderCount-- > 0) {
        auto s = new SenderClient<Ctx>(ctx, senderCount, msgPerSec);
        senders.emplace_back(s);
        ctx.start(*s, cpuMask, true);
        cpuMask <<= 1ul;
    }
    ctx.start();
    os::HandleSignals::onTermIntDo(
        [&ctx]() {
        ctx.stop();
    });

    cout << "ctrl-c to stop" << endl;
    ctx.join();

    for (auto r : receivers) {
        r->showHist();
    }
}

template <typename Ctx>
void runIpc(bool sender, bool receiver) {
    uint64_t cpuMask = 1ul << startCore;
    SingletonGuardian<HMBDC_LOGGER> logGuard(std::cout);
    Ctx ctx("hmbdcperf", bufferSizePower2);
    //its is impossible the receiver not reading anything for 2 sec in this app 
    //the default value is 60 sec
    ctx.setSecondsBetweenPurge(2);

    typename SenderClient<Ctx>::ptr s;
    typename ReceiverClient<Ctx>::ptr r;
    if (sender == receiver) {//creator
        ctx.start();
    } else if (sender) {
        s.reset(new SenderClient<Ctx>(ctx, 0, msgPerSec));
        ctx.start(*s, cpuMask);
    } else {
        r.reset(new ReceiverClient<Ctx>(ctx, 0));
        ctx.start(*r, cpuMask);
    }

    ctx.start();
    os::HandleSignals::onTermIntDo(
        [&ctx]() {
        ctx.stop();
    });
    
    cout << "ctrl-c to stop" << endl;
    ctx.join();
    if(r) {
        r->showHist();
    }
}

using PartionContext = hmbdc::app::Context<24, hmbdc::app::context_property::partition>;
using BroadcastContext = hmbdc::app::Context<24, hmbdc::app::context_property::broadcast<>>;

using IpcCreatorBroadcastContext = hmbdc::app::Context<24
    , hmbdc::app::context_property::broadcast<>
    , hmbdc::app::context_property::ipc_creator
>;

using IpcAttacherBroadcastContext = hmbdc::app::Context<24
    , hmbdc::app::context_property::broadcast<>
    , hmbdc::app::context_property::ipc_attacher
>;
using IpcReceiverBroadcastContext = IpcAttacherBroadcastContext;
using IpcSenderBroadcastContext = IpcAttacherBroadcastContext;

using IpcCreatorPartitionContext = hmbdc::app::Context<24
    , hmbdc::app::context_property::partition
    , hmbdc::app::context_property::ipc_creator
>;

using IpcAttacherPartitionContext = hmbdc::app::Context<24
    , hmbdc::app::context_property::partition
    , hmbdc::app::context_property::ipc_attacher
>;
using IpcReceiverPartitionContext = IpcAttacherPartitionContext;
using IpcSenderPartitionContext = IpcAttacherPartitionContext;


int main(int argc, char** argv) {
    using namespace std;

    namespace po = boost::program_options;
    po::options_description desc("Allowed options");
    auto helpStr = 
    "This program can be used to test hmbdc-free thread and ipc messaging latency performance among a group of cores."
    "sender cores send messages at a specified rate and receiver cores receive and measure message latencies\n"
    "Note: if the rate specified is too high, the results could be impacted by the message processing speed and not accurate.\n"
    "The source code is in example dir: perf-base-lat.cpp"

    "Example:\nrunning \n$ ./perf-base-lat \n"
    "would start the test using default setting.";
    desc.add_options()
    ("help", helpStr)
    ("partition", po::value<bool>(&::part)->default_value(false), "test using partion context (vs the default broadcast context)")
    ("msgPerSec", po::value<size_t>(&msgPerSec)->default_value(0), "the sender sending messages per second - 0 means no rate control")
    ("bufferSizePower2,b", po::value<uint16_t>(&bufferSizePower2)->default_value(1u), "2^bufferSizePower2 is the message buffer size")
    ("startCore,C", po::value<uint16_t>(&startCore)->default_value(0u), "specify starting core number")
    ("skipFirst", po::value<size_t>(&skipFirst)->default_value(1u), "skipp the first N results (warming up stage) when collecting latency stats")
    ("senderCount,s", po::value<int>(&senderCount)->default_value(1u), "how many senders, 1 sender takes up a core ...")
    ("receiverCount,r", po::value<int>(&receiverCount)->default_value(1u), "how many receivers, 1 receiver take up a core")
    ("ipc", po::value<char>(&ipcRole)->default_value('n'), 
        "testing ipc, start one ipc creator (c), ipc sender (s) or ipc receiver (r). "
        "only --partition -b and -C are effective and --partition and -b settings need to be the same when starting "
        "the creator (c), sender (s), and receiver (r). Only one creator is allowed in a test.");

    po::positional_options_description p;
    po::variables_map vm;
    po::store(po::command_line_parser(argc, argv).
        options(desc).positional(p).run(), vm);
    po::notify(vm);

    if (vm.count("help")) {
        cout << desc << "\n";
        return 0;
    }

    switch (ipcRole) {
        case 'n': {
            if (part) run<PartionContext>();
            else run<BroadcastContext>();
        } break;
        case 'c': {
            if (part) runIpc<IpcCreatorPartitionContext>(false, false);
            else runIpc<IpcCreatorBroadcastContext>(false, false);
        } break;
        case 's': {
            if (part) runIpc<IpcSenderPartitionContext>(true, false);
            else runIpc<IpcSenderBroadcastContext>(true, false);
        } break;
        case 'r': {
            if (part) runIpc<IpcReceiverPartitionContext>(false, true);
            else runIpc<IpcReceiverBroadcastContext>(false, true);
        } break;
    }
    return 0;
}
