#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/mcast/Transport.hpp"
#include "hmbdc/app/LoggerT.hpp"
#include "hmbdc/text/StringTrieSet.hpp"
#include "hmbdc//Traits.hpp"

#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>

#include <memory>
#include <type_traits>

namespace hmbdc { namespace app { namespace mcast {

/**
 * @brief interface to power a multicast transport receiving functions
 */
struct RecvTransport : Transport {
    using ptr = std::shared_ptr<RecvTransport>;
    using Transport::Transport;
    virtual ~RecvTransport(){}
    /**
     * @brief a take all arbitrator (no arbitration at all)
     * @details it provides the default type for arbitration which does nothing
     * it also provides a template on writing a user defined arbitrator
     * 
     * @param  h handle to retrieve what is inside of the message
     * @return always returns 1 here
     * In general: 1 keep the message; -1 to drop; 
     * 0 cannot decide, ask me later.
     */
    struct NoOpArb {
        int operator()(TransportMessageHeader const* h) {
            return 1; //always keep it
        }
    };
private:
    friend struct NetContext; //only be used by NetContext
    virtual void listenTo(Topic const& t) = 0;
    virtual void stopListenTo(Topic const& t) = 0;
};

namespace recvtransportengine_detail {
using namespace hmbdc::time;
using namespace boost::asio;

/**
 * @class RecvTransportImpl<>
 * @brief impl class
 * 
 * @tparam OutputBuffer type of buffer to hold resulting network messages
 * @tparam MsgArbitrator arbitrator to decide drop or keep messages, suited to arbitrate 
 * between different recv transport. By default, keeping all
 */
template <typename OutputBuffer, typename MsgArbitrator = RecvTransport::NoOpArb>
struct RecvTransportImpl 
: RecvTransport
, MessageHandler<RecvTransportImpl<OutputBuffer, MsgArbitrator>, Subscribe, Unsubscribe> {
    using SELF = RecvTransportImpl;
    using MH = MessageHandler<SELF, Subscribe, Unsubscribe>;
    friend struct NetContext; //only be created by NetContext

/**
 * @brief ctor
 * @details io_service could be passed in by user, in this case NO more than two threads should
 * power this io_service instance since that would violate the thread garantee of Client, which
 * is no callbacks are called in parallel
 * 
 * @param cfg specify the details of the mcast transport
 * @param outputBuffer holding the results
 * @param arb arbitrator instance to decide which messages to drop and keep; it supports either
 * raw udp packet (BEFORE topic filtering) or hmbdc udpcast message (AFTER topic filtering) level 
 * arbitration depending on which one of 
 * int operator()(void* bytes, size_t len) or 
 * int operator()(TransportMessageHeader const* header) presents in the arb
 */
    RecvTransportImpl(Config const& cfg
        , OutputBuffer& outputBuffer
        , MsgArbitrator arb = NoOpArb())
	: RecvTransport(cfg)
    , buffer_(std::max(sizeof(MessageWrap<Subscribe>), sizeof(MessageWrap<Unsubscribe>))
        , config_.getExt<uint16_t>("cmdBufferSizePower2"))
    , outputBuffer_(outputBuffer)
    , maxItemSize_(outputBuffer.maxItemSize())
    , endpoint_(ip::address::from_string(cfg.getExt<std::string>("mcastAddr"))
        , cfg.getExt<short>("mcastPort"))
    , pSocket_(nullptr)
    , bufCur_(nullptr)
    , bytesRecved_(0)
    , arb_(arb) {
    }

    ~RecvTransportImpl() {
        delete pSocket_;
    }

/**
 * @brief start the show by schedule the mesage recv
 */
    void start() {
        pSocket_->async_receive_from(
            boost::asio::buffer(buf_, sizeof(buf_))
            , sender_endpoint_
            , boost::bind(&SELF::handleReceiveFrom, this,
                  boost::asio::placeholders::error,
                  boost::asio::placeholders::bytes_transferred)
        );
    }

/**
 * @brief power the io_service and other things
 * 
 */
    void runOnce() HMBDC_RESTRICT {
        // if (pIos_->stopped()) pIos_->reset();
        typename Buffer::iterator begin, end;
        auto n = buffer_.peek(0, begin, end);
        auto it = begin;
        while (it != end) {
            MH::handleMessage(*static_cast<MessageHead*>(*it++));
        }
        buffer_.wasteAfterPeek(0, n);
        boost::system::error_code ec;
        if (hmbdc_unlikely(bufCur_)) { //there is unfinished data in buf - arb_
            handleReceiveFrom(ec, 0);
        }
        if (hmbdc_unlikely(pIos_->stopped())) pIos_->reset();
        pIos_->poll(ec);
        if (ec) HMBDC_LOG_C("io_service error=", ec);
    }
   
/**
 * @brief only used by MH
 */
    void handleMessageCb(Subscribe const& t) {
        subscriptions_.add(boost::lexical_cast<std::string>(t.topic));
    }

/**
 * @brief only used by MH
 */
    void handleMessageCb(Unsubscribe const& t) {
        subscriptions_.erase(boost::lexical_cast<std::string>(t.topic));
    }


    void listenTo(Topic const& t) override {
        buffer_.put(MessageWrap<Subscribe>{t});
    }

    void stopListenTo(Topic const& t) override {
        buffer_.put(MessageWrap<Unsubscribe>{t});
    }

    /**
     * @brief expose so user can manipulate it
     * @return reference to boost::asio::ip::udp::socket
     */
    boost::asio::ip::udp::socket& asioSocket() {
        return *pSocket_;
    }

    void initInThread() {
        Transport::initInThread();

        delete pSocket_;
        pSocket_ = new boost::asio::ip::udp::socket(*pIos_);
        pSocket_->open(endpoint_.protocol());
        pSocket_->set_option(ip::udp::socket::reuse_address(true));
        pSocket_->bind(endpoint_); 
        // Join the multicast group.
        auto multicastAddress = 
            ip::address::from_string(config_.getExt<std::string>("mcastAddr"));
        auto listenAddress = ip::address::from_string(
            comm::inet::getLocalIpMatchMask(config_.getExt<std::string>("ifaceAddr")));


        auto sz = config_.getExt<int>("udpRecvBufferBytes");
        if (sz) {
            socket_base::receive_buffer_size option(sz);
            boost::system::error_code ec;
            pSocket_->set_option(option);
        }

        socket_base::receive_buffer_size option;
        pSocket_->get_option(option);
        if (sz == 0 || sz >= option.value()) {
        } else {
            HMBDC_LOG_C("set udp receive buffer size unsuccessful, want "
                , sz, " actual: ", option.value()
                , " resulting lower receiving rate, check OS limits!");
        }
        pSocket_->set_option(ip::multicast::join_group(
            multicastAddress.to_v4(), listenAddress.to_v4())
        );
    }

private:
    template <bool is_raw_arb>
    typename enable_if<is_raw_arb, int>::type applyRawArb() {
        return arb_(bufCur_, bytesRecved_);
    }

    template <bool is_raw_arb>
    typename enable_if<!is_raw_arb, int>::type applyRawArb() {
        return 1;
    }

    template <bool is_raw_arb>
    typename enable_if<!is_raw_arb, int>::type applyArb(TransportMessageHeader* h) {
        return arb_(h);
    }

    template <bool is_raw_arb>
    typename enable_if<is_raw_arb, int>::type applyArb(TransportMessageHeader* h) {
        return 1;
    }

    void handleReceiveFrom(const boost::system::error_code& error
        , size_t bytesRecved) {
        if (hmbdc_likely(!error)) {
            if (!bufCur_) {
                bufCur_ = buf_;
                bytesRecved_ = bytesRecved;
            }
            using traits = 
                pattern::function_traits<typename std::remove_reference<MsgArbitrator>::type>;
            using arg0 = typename traits::template arg<0>::type;
            bool const is_raw_arb = std::is_same<void*, typename std::decay<arg0>::type>::value;
            auto a = applyRawArb<is_raw_arb>();
            if (hmbdc_unlikely(a == 0)) {
                //keep bufCur_ and bytesRecved_ unchanged
                return;
            } else if (hmbdc_unlikely(a < 0)) { //drop the whole packet
                bytesRecved_ = 0;
            } // else process it

            while (bytesRecved_) {
                auto h = reinterpret_cast<TransportMessageHeader*>(bufCur_);
                auto wireSize = h->wireSize();

                if (hmbdc_likely(bytesRecved_ >= wireSize)) {
                    if (subscriptions_.check(h->topic())) {
                        // auto tmp = *(MessageHead*)(h->payload()); HMBDC_LOG_DEBUG(tmp);
                        auto l = std::min<size_t>(maxItemSize_, wireSize);
                    
                        if (hmbdc_unlikely(l > bytesRecved_)) {
                            break;
                        }
                        auto a = applyArb<is_raw_arb>(h);

                        if (hmbdc_likely(a > 0)) {
                            auto it = outputBuffer_.claim();
                            char* b = static_cast<char*>(*it);
                            memcpy(b, h->payload(), l);
                            outputBuffer_.commit(it);
                        } else if (a == 0) {
                            //keep bufCur_ and bytesRecved_ unchanged
                            return;
                        } //else drop and move on
                    }
                    bytesRecved_ -= wireSize;
                    bufCur_ += wireSize;
                } else {
                    break;
                }
            }
        } else {
            HMBDC_LOG_C("asio error=", error);
        }
        bufCur_ = nullptr;
        pSocket_->async_receive_from(
            boost::asio::buffer(buf_, sizeof(buf_))
            , sender_endpoint_
            , boost::bind(&SELF::handleReceiveFrom, this,
                  boost::asio::placeholders::error,
                  boost::asio::placeholders::bytes_transferred)
        );
    }

    Buffer buffer_;
    OutputBuffer& outputBuffer_;
    size_t maxItemSize_;
    boost::asio::ip::udp::endpoint endpoint_;
    boost::asio::ip::udp::endpoint sender_endpoint_;
    ip::udp::socket* pSocket_;
    char buf_[4*1024];
    char* bufCur_;
    size_t bytesRecved_;
    text::StringTrieSet subscriptions_;
    MsgArbitrator arb_;
};


/**
 * @class RecvTransportEngineImpl<>
 * @brief impl class
 * @details this needs to be created using NetContext and start in an app::Context
 * 
 * @tparam OutputBuffer type of buffer to hold resulting network messages
 * @tparam MsgArbitrator arbitrator to decide drop or keep messages, sued to arbitrate 
 * between difefrent recv transports. By default, keep all
 */
template <typename OutputBuffer, typename MsgArbitrator = RecvTransport::NoOpArb>
struct RecvTransportEngineImpl 
: RecvTransportImpl<OutputBuffer, MsgArbitrator>
, single_thread_powered_client
, Client<RecvTransportEngineImpl<OutputBuffer, MsgArbitrator>> {
    using RecvTransportImpl<OutputBuffer, MsgArbitrator>::RecvTransportImpl;


    char const* hmbdcName() const { 
        return this->hmbdcName_.c_str();
    }

    std::tuple<char const*, int> schedSpec() const {
        return std::make_tuple(this->schedPolicy_.c_str(), this->schedPriority_);
    }

/**
 * @brief power the io_service and other things
 * 
 */
    /*virtual*/ 
    void invokedCb(uint16_t) HMBDC_RESTRICT override  {
        RecvTransportImpl<OutputBuffer, MsgArbitrator>::runOnce();
    }


/**
 * @brief start the show by schedule the mesage recv
 */
    /*virtual*/ 
    void messageDispatchingStartedCb(uint16_t threadSerialNumber) override {
        this->initInThread();
        RecvTransportImpl<OutputBuffer, MsgArbitrator>::start();
    };
    
/**
 * @brief should not happen ever unless an exception thrown
 * 
 * @param e exception thown
 */
    /*virtual*/ 
    void stoppedCb(std::exception const& e) override {
        HMBDC_LOG_C(e.what());
    };

};

} //recvtransportengine_detail

template <typename OutputBuffer, typename MsgArbitrator = RecvTransport::NoOpArb>
using RecvTransportImpl = recvtransportengine_detail::RecvTransportImpl<OutputBuffer, MsgArbitrator>;

template <typename OutputBuffer, typename MsgArbitrator = RecvTransport::NoOpArb>
using RecvTransportEngine = recvtransportengine_detail::RecvTransportEngineImpl<OutputBuffer, MsgArbitrator>;
 
}}}
