#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/mcast/Transport.hpp"
#include "hmbdc/app/mcast/Messages.hpp"
#include "hmbdc/app/Base.hpp"
#include "hmbdc/time/Time.hpp"
#include "hmbdc/time/Rater.hpp"
#include "hmbdc/numeric/BitMath.hpp"


#include <boost/bind.hpp>
#include <memory>

#include <iostream>

namespace hmbdc { namespace app { namespace mcast {

namespace sendtransportengine_detail {
using namespace std;    
using namespace hmbdc::time;
using namespace boost::asio;
using namespace hmbdc::pattern;

struct SendTransport 
: Transport {
    using ptr = std::shared_ptr<SendTransport>;
    SendTransport(Config const& cfg
        , size_t maxMessageSize) 
    : Transport(cfg)
    , topic_(cfg.getExt<string>("topicRegex"))
    , topicRegex_(topic_)
    , maxMessageSize_(maxMessageSize)
    , buffer_(maxMessageSize + sizeof(TransportMessageHeader) + sizeof(Topic) + sizeof(MessageHead), outBufferSizePower2())
    , endpoint_(ip::address::from_string(config_.getExt<string>("mcastAddr"))
        , cfg.getExt<short>("mcastPort"))
    , pSocket_(nullptr)
    , rater_(Duration::seconds(1u)
        , config_.getExt<size_t>("sendBytesPerSec")
        , config_.getExt<size_t>("sendBytesBurst")
        , config_.getExt<size_t>("sendBytesBurst") != 0ul) //no rate control by default
    , mtu_(config_.getExt<size_t>("mtu"))
    , maxSendBatch_(config_.getExt<size_t>("maxSendBatch")) {
        mtu_ -= (8u + 20u);  // 8bytes udp header and 20bytes ip header
        auto totalHead = sizeof(app::MessageHead) + sizeof(Topic) + sizeof(TransportMessageHeader);
        if (maxMessageSize_ +  totalHead > mtu_) {
            HMBDC_THROW(std::out_of_range, "maxMessageSize needs <= " << mtu_ - totalHead);
        }
        toSend_.reserve(maxSendBatch_);
    }

    ~SendTransport() {
        delete pSocket_;
    }

    bool match(Topic const& t) const {
        return boost::regex_match(t.c_str(), topicRegex_);
    }

    template <typename... Messages>
    void queue(Topic const& t, Messages&&... msgs) {
        auto n = sizeof...(msgs);
        auto it = buffer_.claim(n);
        queue(it, t, std::forward<Messages>(msgs)...);
        buffer_.commit(it, n);
    }

    template <typename... Messages>
    bool tryQueue(Topic const& t, Messages&&... msgs) {
        auto n = sizeof...(msgs);
        auto it = buffer_.tryClaim(n);
        if (it) {
            queue(it, t, std::forward<Messages>(msgs)...);
            buffer_.commit(it, n);
            return true;
        }
        return false;
    }

    template <typename Message, typename ... Args>
    void queueInPlace(Topic const& t, Args&&... args) {
        auto s = buffer_.claim();
        char* addr = static_cast<char*>(*s);
        auto h = reinterpret_cast<TransportMessageHeader*>(addr);
        auto tl = t.copyTo(addr + sizeof(TransportMessageHeader));
        h->topicLen = tl;
        if (hmbdc_likely(sizeof(Message) <= maxMessageSize_)) {
            new (addr + sizeof(TransportMessageHeader) + tl) MessageWrap<Message>(std::forward<Args>(args)...);
            h->messagePayloadLen() = sizeof(MessageWrap<Message>);
        } else {
            HMBDC_THROW(std::out_of_range
                , "maxMessageSize too small to hold a message when constructing SendTransportEngine");
        }
        buffer_.commit(s);
    }

    void queueBytes(Topic const& t, uint16_t tag, void const* bytes, size_t len) {
        auto s = buffer_.claim();
        char* addr = static_cast<char*>(*s);
        auto h = reinterpret_cast<TransportMessageHeader*>(addr);
        auto tl = t.copyTo(addr + sizeof(TransportMessageHeader));
        h->topicLen = tl;
        if (hmbdc_likely(len <= maxMessageSize_)) {
            new (addr + sizeof(TransportMessageHeader) + tl) MessageWrap<JustBytes>(tag, bytes, len);
            h->messagePayloadLen() = sizeof(MessageWrap<JustBytes>) - sizeof(MessageWrap<JustBytes>::payload) + len;
        } else {
            HMBDC_THROW(std::out_of_range
                , "maxMessageSize too small to hold a message when constructing SendTransportEngine");
        }
        buffer_.commit(s);

    }

    void runOnce() HMBDC_RESTRICT {
        if (!toSend_.size()) {
            if (hmbdc_unlikely(pIos_->stopped())) pIos_->reset();
            boost::system::error_code error;
            size_t bytesTransferred = 0;
            resumeSend(error, bytesTransferred);
        }
        boost::system::error_code ec;
        pIos_->poll(ec);
        if (ec) HMBDC_LOG_C("io_service error=", ec);
    }

    void stop() {
        buffer_.reset(0);
    }

    /**
     * @brief expose so user can manipulate it
     * @return reference to boost::asio::ip::udp::socket
     */
    boost::asio::ip::udp::socket& asioSocket() {
        return *pSocket_;
    }

    void initInThread() {
        Transport::initInThread();
        delete pSocket_;
        pSocket_ = new boost::asio::ip::udp::socket(*pIos_, endpoint_.protocol());

        auto iface = 
            comm::inet::getLocalIpMatchMask(config_.getExt<string>("ifaceAddr"));
        ip::multicast::outbound_interface outFrom(ip::address_v4::from_string(iface));
        pSocket_->set_option(outFrom);

        ip::multicast::hops hops(config_.getExt<uint16_t>("ttl"));
        pSocket_->set_option(hops);

        // uint8_t loopch = config_.getExt("loopback", false)?1:0;
        // if (setsockopt(pSocket_->native_handle(), IPPROTO_IP, IP_MULTICAST_LOOP
        //     , (char *)&loopch, sizeof(loopch) < 0)) {
        //     HMBDC_LOG_C("cannot set IP_MULTICAST_LOOP to ", (int)loopch);
        // }

        ip::multicast::enable_loopback loop(config_.getExt<bool>("loopback"));
        pSocket_->set_option(loop);

        auto sz = config_.getExt<int32_t>("udpSendBufferBytes");
        if (sz) {
            socket_base::send_buffer_size option(sz);
            boost::system::error_code ec;
            pSocket_->set_option(option);
        }

        socket_base::send_buffer_size option;
        pSocket_->get_option(option);
        if (sz == 0 || sz >= option.value()) {
        } else {
            HMBDC_LOG_C("set udp send buffer size unsuccessful, want "
                , sz, " actual: ", option.value());
        }
        //we need to wait for mcast registration to settle, otherwise sent msg might drop
        usleep(10000);
    }

private:
    std::string topic_;
    boost::regex topicRegex_;  
    size_t maxMessageSize_;
    Buffer buffer_;
    typename Buffer::iterator begin_;

    ip::udp::endpoint endpoint_;  
    ip::udp::socket* pSocket_;
    Rater rater_;
    size_t mtu_;
    size_t maxSendBatch_;
    vector<const_buffer> toSend_;


    uint16_t
    outBufferSizePower2() {
        auto res = config_.getExt<uint16_t>("outBufferSizePower2");
        if (res) {
            return res;
        }
        res =hmbdc::numeric::log2Upper(8ul * 1024ul / (8ul + maxMessageSize_));
        HMBDC_LOG_N("auto set --outBufferSizePower2=", res);
        return res;
    }

    template<typename M, typename ... Messages>
    void queue(typename Buffer::iterator it, Topic const& t
        , M&& m, Messages&&... msgs) {
        using Message = typename std::remove_reference<M>::type;
        auto s = *it;
        char* addr = static_cast<char*>(s);
        auto h = reinterpret_cast<TransportMessageHeader*>(addr);
        auto tl = t.copyTo(addr + sizeof(TransportMessageHeader));
        h->topicLen = tl;
        if (hmbdc_likely(sizeof(Message) <= maxMessageSize_)) {
            new (addr + sizeof(TransportMessageHeader) + tl) MessageWrap<Message>(std::forward<M>(m));
            h->messagePayloadLen() = sizeof(MessageWrap<Message>);
        } else {
            HMBDC_THROW(std::out_of_range
                , "maxMessageSize too small to hold a message when constructing SendTransportEngine");
        }
        queue(++it, t, std::forward<Messages>(msgs)...);
    }

    void queue(typename Buffer::iterator it
        , Topic const& t) {} 

    void resumeSend(const boost::system::error_code& error
        , size_t bytesTransferred) {
        if (hmbdc_unlikely(error)) {
            HMBDC_LOG_C("asio error=", error);
        }
        if (hmbdc_likely(begin_)) 
            buffer_.wasteAfterPeek(0, toSend_.size());
        toSend_.clear();

        typename Buffer::iterator end;
        buffer_.peek(0, begin_, end, maxSendBatch_);
        auto it = begin_;
        size_t batchBytes = 0;
        while (it != end) {
            void* ptr = *it;
            auto item = static_cast<TransportMessageHeader*>(ptr);
            if (hmbdc_unlikely(!rater_.check(item->wireSize()))) break;
            batchBytes += item->wireSize();
            if (hmbdc_unlikely(batchBytes > mtu_)) break;
            it++;
            toSend_.emplace_back(ptr, item->wireSize()); 
            rater_.commit();
        }

        if (hmbdc_unlikely(toSend_.size() == 1)) {
            pSocket_->send_to(toSend_, endpoint_);
            buffer_.wasteAfterPeek(0, 1);
            toSend_.clear();
            begin_.clear();
        } else if (hmbdc_likely(toSend_.size())) {
            pSocket_->async_send_to(
                toSend_
                , endpoint_
                , boost::bind(&SendTransport::resumeSend
                    , this
                    , boost::asio::placeholders::error
                    , boost::asio::placeholders::bytes_transferred)
            );
        }
    }
};

struct SendTransportEngine
: SendTransport
, single_thread_powered_client
, Client<SendTransportEngine> {  
    using SendTransport::SendTransport;


    char const* hmbdcName() const { 
        return this->hmbdcName_.c_str();
    }

    std::tuple<char const*, int> schedSpec() const {
        return std::make_tuple(this->schedPolicy_.c_str(), this->schedPriority_);
    }

    /*virtual*/ 
    void messageDispatchingStartedCb(uint16_t threadSerialNumber) override {
        this->initInThread();
    }

    /*virtual*/ 
    void invokedCb(uint16_t threadSerialNumber) HMBDC_RESTRICT override  {
        runOnce();
    }

    /*virtual*/ bool droppedCb() override {
        stop();
        return true;
    };
};

} //sendtransportengine_detail
using SendTransport = sendtransportengine_detail::SendTransport;
using SendTransportEngine = sendtransportengine_detail::SendTransportEngine;
}}}
