#include "hmbdc/Copyright.hpp"
#pragma once 
#include "hmbdc/app/mcast/SendTransportEngine.hpp"

#include <type_traits>
#include <memory>

namespace hmbdc { namespace app { namespace mcast {
struct NetContext;
/**
 * @brief fascade class for sending network messages
 */

struct Sender {
	using ptr = std::shared_ptr<Sender>;

private:
    friend struct hmbdc::app::mcast::NetContext;
	Sender(SendTransport::ptr transport, Topic const& t)
	: transport_(transport)
	, topic_(t) {
	}

public:
    
    /**
     * @brief send a message's first bytes
     * @details the bytes length could be larger than Message, but needs to 
     * be able to fit in the tansport buffer the sender is associated 
     * with - exception thrown otherwise
     * 
     * @param msg the message to send
     * @tparam Message Type
     * @param len just send  the first len bytes of this message
     * @tparam T integral type
     */
    template <typename Message, typename T
        , typename = std::enable_if<std::is_integral<T>::value>>
    void send(Message&& msg, T len) {
        using raw = typename std::remove_reference<Message>::type;
        transport_->queueBytes(
            topic_, raw::typeTag, &msg, static_cast<size_t>(len));
    }

    /**
     * @brief send a batch of message asynchronizely 
     * @details although batching could implicitly happen
     *  further performance gain could be achieved by sending more msg in 
     *  a batch here
     * 
     * @param m Message
     * @tparam Message Type
     */
    template <typename... Messages>
    void send(Messages&&... msgs) {
        transport_->queue(topic_, std::forward<Messages>(msgs)...);
    }
    
    /**
     * @brief send a batch of message asynchronizely
     * @details this call does not block and is transactional - all or none is queued
     * 
     * @param m Message
     * @tparam Message Type
     * 
     * @return true if messages are queued successfully
     */
    template <typename... Messages>
    bool trySend(Messages&&... msgs) {
        return transport_->tryQueue(topic_, std::forward<Messages>(msgs)...);
    }

    /**
     * @brief send a message asynchronizely - avoiding Message copying
     * by directly constructing the message in the buffer
     * 
     * @param args Message's ctor args to construct the Message in the buffer
     * @tparam Message Type
     * @tparam typename ... Args args type
     */
    template <typename Message, typename ... Args>
    void sendInPlace(Args&&... args) {
    	transport_->template queueInPlace<Message>(topic_, std::forward<Args>(args)...);
    }

    /**
     * @brief send a message asynchronizely by providing message 
     * in tag and bytes 
     * @details for runtime typed usage when  the message type can only be decided
     * at runtime
     * 
     * @param tag message tag
     * @param bytes message byte starting address
     * @param len byte length of the above
     */
    void sendBytes(uint16_t tag, void const* bytes, size_t len) {
        transport_->queueBytes(topic_, tag, bytes, len);
    } 

private:
	SendTransport::ptr transport_;
	comm::Topic topic_;
};
}}}
