#include "hmbdc/Copyright.hpp"
#pragma once 

#include "hmbdc/comm/Topic.hpp"
#include "hmbdc/time/Time.hpp"
#include "hmbdc/app/Message.hpp"

#include <iostream>
#include <string>

#include <unistd.h>

namespace hmbdc { namespace app { namespace tcpcast {

struct TransportMessageHeader {
    uint8_t topicLen;
    uint8_t flag; //1 - hasAttachment
    uint16_t messagePayloadLen;

    std::pair<char const*, char const*> topic() const {
        char const* b = reinterpret_cast<const char*>(this) 
            + sizeof(TransportMessageHeader); 
        return std::make_pair(b, b + topicLen);
    }

    void const* payload() const {
        return reinterpret_cast<const char*>(this) 
            + sizeof(TransportMessageHeader) + topicLen; 
    }

    void * payload() {
        return reinterpret_cast<char*>(this) 
            + sizeof(TransportMessageHeader) + topicLen; 
    }

    uint16_t typeTag() const {
        auto h = static_cast<app::MessageHead const*>(payload());
        return h->typeTag;
    }

    template <typename Message>
    Message& wrapped() {
        auto wrap = static_cast<app::MessageWrap<Message>*>(payload());
        return wrap->payload;
    }

    template <typename Message>
    Message const& wrapped() const {
        auto wrap = static_cast<app::MessageWrap<Message> const *>(payload());
        return wrap->payload;
    }

    size_t wireSize() const {
        return sizeof(TransportMessageHeader) + topicLen 
            + messagePayloadLen;
    }    

    size_t wireSizeContainsTopic() const {
        return sizeof(TransportMessageHeader) + topicLen;
    }
} __attribute__((packed));

struct Subscribe
: hasTag<201> {
    Subscribe(comm::Topic const&topic)
    : topic(topic){}
    comm::Topic topic;
};

struct Unsubscribe 
: hasTag<202> {
    Unsubscribe(comm::Topic const&topic)
    : topic(topic){}
    comm::Topic topic;
};

struct TopicSource 
: hasTag<203> {
    TopicSource()
    : pid(0)
    , loopback(false)
    , timestamp(time::SysTime::now())
    {}

    TopicSource(std::string const&topicRegexIn
        , std::string const& ipIn
        , uint16_t portIn
        , bool loopbackIn)
    : port(portIn)
    , pid(getpid())
    , loopback(loopbackIn)
    , timestamp(time::SysTime::now()) {
        strncpy(ip, ipIn.c_str(), sizeof(ip));
        strncpy(topicRegex, topicRegexIn.c_str(), sizeof(topicRegex));
        topicRegex[sizeof(topicRegex) - 1] = 0;
    }
    char ip[16];
    char topicRegex[128]; //null terminated string
    uint16_t port;
    pid_t pid;
    bool loopback;
    time::SysTime timestamp; //when the source started

    friend
    std::ostream& operator << (std::ostream& os, TopicSource const& m) {
        return os << m.ip << ' ' 
            << m.port << ' ' 
            << m.topicRegex;
    }

    friend
    std::istream& operator >> (std::istream& is, TopicSource& m) {
        is 
            >> m.ip 
            >> m.port 
            >> m.topicRegex
            ;
        return is;
    }
};

/**
 * @class SessionStarted
 * @brief this message appears in the receiver's buffer indicating a new source is connected
 * @details only appears on the receiving side, and the receiver buffer is big enough to hold this messages
 */
struct SessionStarted 
: hasTag<204> {
    char ip[16];
    friend 
    std::ostream& operator << (std::ostream& os, SessionStarted const & m) {
        return os << "Session to topic source started " << m.ip;
    }
};

/**
 * @class SessionDropped
 * @brief this message appears in the receiver's buffer indicating a previously connected source is dropped
 * @details only appears on the receiving side, and the receiver buffer is big enough to hold this messages
 */
struct SessionDropped 
: hasTag<205> {
    char ip[16];
    friend 
    std::ostream& operator << (std::ostream& os, SessionDropped const & m) {
        return os << "Session to topic source dropped " << m.ip;
    }
};

}}}
