#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/tcpcast/Transport.hpp"
#include "hmbdc/app/tcpcast/RecvSession.hpp"
#include "hmbdc/app/tcpcast/Messages.hpp"
#include "hmbdc/app/mcast/RecvTransportEngine.hpp"
#include "hmbdc/app/LoggerT.hpp"
#include "hmbdc/text/StringTrieSet.hpp"

#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/unordered_map.hpp>

#include <memory>

namespace hmbdc { namespace app { namespace tcpcast {

/**
 * @brief interface to power a tcpcast transport receiving functions
 */
struct RecvTransport 
: Transport {
    using ptr = std::shared_ptr<RecvTransport>;
    using Transport::Transport;
    virtual ~RecvTransport(){}
    /**
     * @brief a take all arbitrator (no arbitration at all)
     * @details it provides the default type for arbitration which does nothing
     * it also provides a template on writing a user defined arbitrator
     * 
     * @param  h handle to retrieve what is inside of the message
     * @return always returns 1 here
     * In general: 1 keep the message; -1 to drop; 
     * 0 cannot decide, ask me later.
     */
    struct NoOpArb {
        int operator()(TransportMessageHeader const* h) {
            return 1; //always keep it
        }
    };

private:
    friend struct NetContext; //only be used by NetContext
    virtual void listenTo(comm::Topic const& t) = 0;
    virtual void stopListenTo(comm::Topic const& t) = 0;
};


namespace recvtransportengine_detail {
using namespace hmbdc::time;
using namespace boost::asio;
using boost::asio::ip::tcp;
using namespace std;
/**
 * @brief impl class
 * @details this needs to be created using NetContext and start in an app::Context
 * 
 * @tparam OutputBuffer type of buffer to hold resulting network messages
 */
template <typename OutputBuffer, typename MsgArbitrator>
struct RecvTransportEngine 
: RecvTransport
, TimerManager
, ReoccuringTimer
, single_thread_powered_client
, Client<RecvTransportEngine<OutputBuffer, MsgArbitrator>>
, MessageHandler<RecvTransportEngine<OutputBuffer, MsgArbitrator>
    , Subscribe, Unsubscribe, TopicSource> {
    using SELF = RecvTransportEngine;
    using MH = MessageHandler<SELF, Subscribe, Unsubscribe, TopicSource>;
    friend struct NetContext; //only be created by NetContext

/**
 * @brief ctor
 * @details io_service could be passed in by user, in this case NO more than two threads should
 * power this io_service instance since that would violate the thread garantee of Client, which
 * is no callbacks are called in parallel
 * 
 * @param cfg specify the details of the tcpcast transport
 * @param outputBuffer holding the results
 * @param arb arbitrator instance to decide which messages to drop and keep; it ONLY supports 
 * hmbdc message (AFTER topic filtering) level (NO packet level since it is tcp)
 */
    RecvTransportEngine(Config const& cfg
        , OutputBuffer& outputBuffer
        , MsgArbitrator arb = NoOpArb())
    : RecvTransport(cfg)
    , ReoccuringTimer(Duration::seconds(cfg.getExt<size_t>("heartbeatPeriodSeconds")))
    , config_(cfg)
    , mcConfig_(cfg) 
    , buffer_(std::max({sizeof(MessageWrap<Subscribe>)
        , sizeof(MessageWrap<Unsubscribe>)
        , sizeof(MessageWrap<TopicSource>)
        }), config_.getExt<uint16_t>("cmdBufferSizePower2"))
    , outputBuffer_(outputBuffer)
    , maxItemSize_(outputBuffer.maxItemSize())
    // , mcReceiver_(cfg, buffer_, &*pIos_)
    , arb_(arb) 
    , myIp_(ip::address_v4::from_string(
        hmbdc::comm::inet::getLocalIpMatchMask(cfg.getExt<string>("ifaceAddr"))
        ).to_ulong())
    , additionalTopicSourcesScanTimer_(
        Duration::seconds(cfg.getExt<size_t>("additionalTopicSourcesScanPeriodSeconds"))) {
        if (outputBuffer.maxItemSize() < sizeof(SessionStarted) 
            || outputBuffer.maxItemSize() < sizeof(SessionDropped))  {
            HMBDC_THROW(std::out_of_range
                , "buffer is too small for notification: SessionStarted and SessionDropped");
        }
        
        // mcConfig_.put("ifaceAddr", cfg.getExt<string>("ifaceAddr"));
        mcConfig_.put("outBufferSizePower2", 3u);
        // mcConfig_.put("mcastPort", cfg.getExt<uint16_t>("mcastPort"));
        // mcConfig_.put("mcastAddr", cfg.getExt<string>("mcastAddr"));

        mcReceiver_.reset(new mcast::RecvTransportImpl<decltype(buffer_)>(
            mcConfig_, buffer_, mcast::RecvTransport::NoOpArb()));

        cfg(additionalTopicSources_, "additionalTopicSources");
    }

/**
 * @brief in the environment that multicast cannot be enabled on either the sender or the receiver side,
 * the sender info cannot be automatically detected by the receiver. 
 * User can use this method to explicitedly tell the receiver about the sender information
 * @details not like additionalTopicSources config which will periodically retry, 
 * this call has no effect if the sender is not up and fully ready
 * 
 * @param s describing the sender
 */
    void tryTopicSource(TopicSource const& s) {
        buffer_.put(MessageWrap<TopicSource>(s));
    }

/**
 * @brief start the show by schedule the message recv
 */
    /*virtual*/ 
    void messageDispatchingStartedCb(uint16_t threadSerialNumber) override {
        this->initInThread();
        mcReceiver_->initInThread();
        mcReceiver_->start();
        mcReceiver_->listenTo(tcpcastAdTopic_);

        setCallback(
            [this](TimerManager& tm, SysTime const& now) {
                for (auto& s : recvSessions_) {
                    s.second->heartbeat();
                }
            }
        );

        schedule(SysTime::now(), *this);
        if (additionalTopicSources_.size()) {
            additionalTopicSourcesScanTimer_.setCallback(
                [this](TimerManager& tm, SysTime const& now) {
                    for (auto const& s : additionalTopicSources_) {
                         handleMessageCb(s);
                    }
                }
            );
            schedule(SysTime::now(), additionalTopicSourcesScanTimer_);
        }
    };

    char const* hmbdcName() const { 
        return this->hmbdcName_.c_str();
    }

    std::tuple<char const*, int> schedSpec() const {
        return std::make_tuple(this->schedPolicy_.c_str(), this->schedPriority_);
    }

/**
 * @brief should not happen ever unless an exception thrown
 * 
 * @param e exception thown
 */
    /*virtual*/ 
    void stoppedCb(std::exception const& e) override {
        HMBDC_LOG_C(e.what());
    };

/**
 * @brief power the io_service and other things
 * 
 */
    /*virtual*/ 
    void invokedCb(uint16_t threadSerialNumber) HMBDC_RESTRICT override  {
        pattern::MonoLockFreeBuffer::iterator begin, end;
        auto n = buffer_.peek(begin, end);
        auto it = begin;
        while (it != end) {
            MH::handleMessage(*static_cast<MessageHead*>(*it++));
        }
        buffer_.wasteAfterPeek(begin, n);
        mcReceiver_->runOnce();
    }
   
/**
 * @brief only used by MH
 */
    void handleMessageCb(Subscribe const& t) {
        subscriptions_.add(boost::lexical_cast<std::string>(t.topic));
        for (auto& p : recvSessions_) {
            p.second->sendSubscribe(t.topic);
        }
    }

/**
 * @brief only used by MH
 */
    void handleMessageCb(Unsubscribe const& t) {
        subscriptions_.erase(boost::lexical_cast<std::string>(t.topic));
        for (auto& p : recvSessions_) {
            p.second->sendUnsubscribe(t.topic);
        }
    }

/**
 * @brief only used by MH
 */
    void handleMessageCb(TopicSource const& t) {
        auto ip = ip::address_v4::from_string(t.ip).to_ulong();
        //unless using loopback address, don't EVER listen to myself (own process)
        if (ip != 2130706433ul //127.0.0.1
            && ip == myIp_ 
            && (!t.loopback || t.pid == getpid())) {
            return;
        }
        auto key = make_pair(ip, t.port);
        if (recvSessions_.find(key) == recvSessions_.end()) {
            boost::regex topicRegex(t.topicRegex);
            for (auto it = subscriptions_.begin(); it != subscriptions_.end(); ++it) {
                if (boost::regex_match(it->first.c_str(), topicRegex)
                    || it->second) { //when wildcard is involved, assuming we need the connection
                    tcp::resolver resolver(*pIos_);
                    auto epIt = resolver.resolve(
                        tcp::resolver::query(t.ip, to_string(t.port)));
                    auto sess = new Session(
                        config_
                        , subscriptions_
                        , topicRegex
                        , [this, key]() {
                            recvSessions_.erase(key);
                        }
                        , *pIos_
                        , outputBuffer_
                        , arb_
                    );
                    recvSessions_[key] = typename Session::ptr(sess);
                    sess->start(epIt);
                    break;
                }
            }
        } //else ignore
    }

private:
    using Session = RecvSession<OutputBuffer, MsgArbitrator>;
    void listenTo(Topic const& t) override {
        buffer_.put(MessageWrap<Subscribe>{t});
    }

    void stopListenTo(Topic const& t) override {
        buffer_.put(MessageWrap<Unsubscribe>{t});
    }
    Config config_, mcConfig_;
    pattern::MonoLockFreeBuffer buffer_;
    OutputBuffer& outputBuffer_;
    size_t maxItemSize_;
    unique_ptr<mcast::RecvTransportImpl<decltype(buffer_)>> mcReceiver_;

    text::StringTrieSet subscriptions_;
    boost::unordered_map<pair<uint64_t, uint16_t>
        , typename Session::ptr> recvSessions_;
    MsgArbitrator arb_;
    uint64_t myIp_;
    vector<TopicSource> additionalTopicSources_;
    ReoccuringTimer additionalTopicSourcesScanTimer_;
};

} //recvtransportengine_detail
template <typename OutputBuffer, typename MsgArbitrator>
using RecvTransportEngine = recvtransportengine_detail::RecvTransportEngine<OutputBuffer, MsgArbitrator>;
}}}
