#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/LoggerT.hpp"
#include "hmbdc/app/tcpcast/Transport.hpp"
#include "hmbdc/app/tcpcast/Messages.hpp"
#include "hmbdc/time/Time.hpp"
#include "hmbdc/pattern/MonoLockFreeBuffer.hpp"
#include "hmbdc/comm/inet/Misc.hpp"

#include "hmbdc/app/tcpcast/SendSession.hpp"

#include <unordered_set>
#include <memory>
#include <utility>
#include <boost/asio.hpp>

#include <iostream>

namespace hmbdc { namespace app { namespace tcpcast {

namespace sendserver_detail {
using namespace hmbdc::pattern;
using namespace boost::asio;
using boost::asio::ip::tcp;
using namespace std;

struct SendServer {
    SendServer(Config const& cfg
        , io_service& ios)
    : config_(cfg)
    , acceptor_(ios
        , tcp::endpoint(
            ip::address::from_string(
                hmbdc::comm::inet::getLocalIpMatchMask(cfg.getExt<string>("ifaceAddr"))
            ), cfg.getExt<short>("tcpPort"))
    )
    , socket_(ios)
    , advertisingMessage_(
        cfg.getExt<string>("topicRegex")
        , acceptor_.local_endpoint().address().to_string()
        , acceptor_.local_endpoint().port()
        , config_.getExt<bool>("loopback")
    ) {
    }

    TopicSource const& advertisingMessage() const {
        return advertisingMessage_;
    }

    virtual size_t readySessionCount() const = 0;
    virtual ~SendServer() = default;
protected:
    Config config_;

    tcp::acceptor acceptor_;
    tcp::socket socket_;
    TopicSource advertisingMessage_;
};

struct AsyncSendServer 
: SendServer {
    AsyncSendServer(Config const& cfg
        , io_service& ios
        , size_t toSendQueueMaxSize)
    : SendServer(cfg, ios) {
        toSendQueue_.set_capacity(toSendQueueMaxSize);
        doAccept();
    }

    void queue(pair<char const*, char const*> const& t
        , ToSend && toSend
        , size_t toSendByteSize
        , MonoLockFreeBuffer::iterator it) {
        toSendQueue_.push_back(forward_as_tuple(t, forward<ToSend>(toSend), toSendByteSize, it));
    }

    /**
     * @brief run the server's async send function and decide which items in buffer
     * can be release
     * @details called all the time
     * 
     * @param begin if nothing can be released in buffer return this
     * @param end if all can be releases return this
     * 
     * @return iterator in buffer poing to new start (not released)
     */
    MonoLockFreeBuffer::iterator run(MonoLockFreeBuffer::iterator begin
        , MonoLockFreeBuffer::iterator end) HMBDC_RESTRICT {
        if (sessions_.size() == 0) {
            toSendQueue_.clear();
            return end;
        }
        //nothing to retire by default
        auto newStartIt = begin;

        auto minIndex = numeric_limits<ToSendQueue::size_type>::max();
        for (auto it = sessions_.begin(); it != sessions_.end();) {
            if (minIndex > (*it)->toSendQueueIndex_) {
                minIndex = (*it)->toSendQueueIndex_;
            }
            if (hmbdc_unlikely(it->use_count() == 1)) {
                (*it)->stop();
                sessions_.erase(it++);
            } else {
                (*it)->run();
                it++;
            }
        }
        if (minIndex) {
            newStartIt = get<3>(toSendQueue_[minIndex - 1]);
            toSendQueue_.erase_begin(minIndex);
            for (auto it = sessions_.begin(); it != sessions_.end(); ++it) {
                (*it)->toSendQueueIndex_ -= minIndex;
            }
        }

        return newStartIt;
    }

    size_t readySessionCount() const override {
        size_t res = 0;
        for (auto const& s : sessions_) {
            if (s->ready()) res++;
        }
        return res;
    }

    void killSlowestSession() {
        for (auto it = sessions_.begin(); it != sessions_.end(); ++it) {
            if (0 == (*it)->toSendQueueIndex_) {
                HMBDC_LOG_C((*it)->id(), " too slow, dropping");
                (*it)->stop();  
                break;
            }
        }
    }

private:
    void doAccept() {
        acceptor_.async_accept(socket_,
            [this](boost::system::error_code ec) {
                if (!ec) {
                    auto sz = config_.getExt<int>("tcpSendBufferBytes");
                    if (sz) {
                        socket_base::send_buffer_size option(sz);
                        boost::system::error_code ec;
                        socket_.set_option(option);
                    }
                    
                    socket_base::send_buffer_size option;
                    socket_.get_option(option);
                    if (sz == 0 || sz >= option.value()) {
                        // HMBDC_LOG_N("tcpcast SendSession send buffer size: ", option.value());
                    } else {
                        HMBDC_LOG_C("set tcpcast SendSession send buffer size unsuccessful, want "
                            , sz, " actual: ", option.value()
                            , " resulting higher receiver dropping possibility, check OS limits!");
                    }

                    socket_.set_option(ip::tcp::no_delay(!config_.getExt<bool>("nagling")));

                    auto s = std::make_shared<AsyncSendSession>(std::move(socket_), toSendQueue_);
                    sessions_.insert(s);
                    s->start();
                }
                doAccept();
            }
        );
    }

    using Sessions = unordered_set<AsyncSendSession::ptr>;
    Sessions sessions_;
    ToSendQueue toSendQueue_; 
};
} //sendserver_detail
using AsyncSendServer = sendserver_detail::AsyncSendServer;
}}}
