#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/LoggerT.hpp"
#include "hmbdc/app/tcpcast/Transport.hpp"
#include "hmbdc/app/tcpcast/Messages.hpp"

#include "hmbdc/text/StringTrieSet.hpp"
#include "hmbdc/time/Time.hpp"
#include "hmbdc/pattern/MonoLockFreeBuffer.hpp"
#include "hmbdc/comm/inet/Misc.hpp"

#include <boost/asio.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/circular_buffer.hpp>

#include <memory>
#include <utility>
#include <iostream>

namespace hmbdc { namespace app { namespace tcpcast {

namespace sendserver_detail {
struct AsyncSendServer;
}

using ToSend = std::vector<boost::asio::const_buffer>;
using ToSendQueue = boost::circular_buffer<std::tuple<std::pair<char const*, char const*> //topic
    , ToSend
    , size_t //total bytes above
    , hmbdc::pattern::MonoLockFreeBuffer::iterator //end iterator
    >
>;

namespace sendsession_detail {
using namespace boost::asio;
using boost::asio::ip::tcp;
using namespace std;



struct SendSession 
: std::enable_shared_from_this<SendSession> {
    SendSession(tcp::socket socket)
    : socket_(move(socket))
    , filledLen_(0)
    , id_(boost::lexical_cast<string>(socket_.remote_endpoint()))
    , ready_(false) {
        socket_.non_blocking(true);
    }

    ~SendSession() {
        HMBDC_LOG_N("SendSession retired: ", id());
    }

    void start() {
        HMBDC_LOG_N("SendSession started: ", id());
        doRead();
    }

    void stop() {
        ready_ = false;
        boost::system::error_code ec;
        socket_.shutdown(tcp::socket::shutdown_both, ec);
        socket_.close();
    }
    
    char const* id() const {
        return id_.c_str();
    }

    bool ready() const {
        return ready_;
    }

private:
    void doRead() {
        while (filledLen_) {
            auto p = find(data_, data_ + filledLen_, '\t');
            if (p != data_ + filledLen_) {
                *p = '\000';
                string t(data_ + 1);
                if (t.size() == 0) {
                    ready_ = true;
                }
                else if (data_[0] == '+') {
                    clientSubscriptions_.add(t);
                } else if (data_[0] == '-') {
                    clientSubscriptions_.erase(t);
                } // else ignore
                memmove(data_, p + 1, filledLen_ - (p - data_ + 1));
                filledLen_ -= p - data_ + 1;
            } else {
                break; //no complete topic request
            }
        }

        auto self(shared_from_this());
        socket_.async_read_some(boost::asio::buffer(data_ + filledLen_, sizeof(data_) - filledLen_),
            [this, self](boost::system::error_code ec, std::size_t length) {
                if (!ec) {
                    filledLen_ += length;
                    doRead();
                }
        });
    }


protected:
    tcp::socket socket_;
    text::StringTrieSet clientSubscriptions_;

private:
    char data_[1024];
    size_t filledLen_;
    string id_;
    bool ready_;

};


struct SyncSendSession
: SendSession {
    using ptr = shared_ptr<SyncSendSession>;
    using SendSession::SendSession;
    template <typename ToSend>
    size_t send(std::pair<char const*, char const*> const& t
        , ToSend const& toSend, size_t size) {
        if (clientSubscriptions_.check(t)) {
            boost::system::error_code ec;
            auto res = write(socket_, toSend, ec);
            if (hmbdc_unlikely(ec)) {
                HMBDC_LOG_W(id(), ", sending error_code=", ec);
                return 0;
            }
            return res;
        }
        return size; //pretend things normal
    }
};

struct AsyncSendSession
: SendSession {
    using ptr = shared_ptr<AsyncSendSession>;
    AsyncSendSession(tcp::socket socket
        , ToSendQueue & toSendQueue)
    : SendSession(move(socket))
    , toSendQueue_(toSendQueue)
    , toSendQueueIndex_(0)
    , asyncWritePending_(false) {
    }

    void run() {
        if (asyncWritePending_) return;

        for (; toSendQueueIndex_ != toSendQueue_.size();) {
            if (clientSubscriptions_.check(get<0>(toSendQueue_[toSendQueueIndex_]))) {
                auto self(shared_from_this());
                auto bytesToSend = get<2>(toSendQueue_[toSendQueueIndex_]);
                async_write(socket_, get<1>(toSendQueue_[toSendQueueIndex_])
                    , transfer_exactly(bytesToSend)
                    , [this, self](const boost::system::error_code& ec, 
                        std::size_t bytes_transferred) {
                        toSendQueueIndex_++;
                        if (hmbdc_unlikely(ec)) {
                            HMBDC_LOG_C(id(), ", async sending error_code=", ec);
                        } else {
                            asyncWritePending_ = false;
                            run();
                        }
                    }
                );
                asyncWritePending_ = true;
                break;
            } else {
                toSendQueueIndex_++;
            }
        }
    }

private:
    friend struct hmbdc::app::tcpcast::sendserver_detail::AsyncSendServer;
    ToSendQueue& toSendQueue_;
    ToSendQueue::size_type toSendQueueIndex_;
    bool asyncWritePending_;
};
} //sendsession_detail

using AsyncSendSession = sendsession_detail::AsyncSendSession;

}}}
