#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/LoggerT.hpp"
#include "hmbdc/app/tcpcast/Transport.hpp"
#include "hmbdc/app/tcpcast/SendServer.hpp"
#include "hmbdc/app/tcpcast/Messages.hpp"
#include "hmbdc/app/mcast/SendTransportEngine.hpp"
#include "hmbdc//Traits.hpp"
#include "hmbdc/time/Time.hpp"
#include "hmbdc/time/Rater.hpp"
#include "hmbdc/numeric/BitMath.hpp"

#include <boost/circular_buffer.hpp>
#include <memory>
#include <tuple>
#include <type_traits>

#include <iostream>

namespace hmbdc { namespace app { namespace tcpcast {

namespace send_detail {
using namespace hmbdc::time;
using namespace boost::asio;
using namespace std;
using pattern::MonoLockFreeBuffer;
/**
 * @brief capture the transportation mechanism
 * 
 */
struct SendTransport 
: Transport {
    using ptr = std::shared_ptr<SendTransport>;
    /**
     * @brief ctor
     * @param cfg jason specifing the transport - see example, perf-tcpcast.cpp
     * @param maxMessageSize max messafe size in bytes to be sent
     */
    SendTransport(Config const& cfg
        , size_t maxMessageSize) 
    : Transport(cfg)
    , topic_(cfg.getExt<string>("topicRegex"))
    , topicRegex_(topic_)
    , maxMessageSize_(maxMessageSize)
    , minRecvToStart_(cfg.getExt<size_t>("minRecvToStart"))
    , buffer_(maxMessageSize + sizeof(TransportMessageHeader) + sizeof(Topic) + sizeof(MessageHead)
        , config_.getExt<uint16_t>("outBufferSizePower2")
            ?config_.getExt<uint16_t>("outBufferSizePower2")
            :hmbdc::numeric::log2Upper(128ul * 1024ul / maxMessageSize)
    ) 
    , rater_(Duration::seconds(1u)
        , config_.getExt<size_t>("sendBytesPerSec")
        , config_.getExt<size_t>("sendBytesBurst")
        , config_.getExt<size_t>("sendBytesBurst") != 0ul //no rate control by default
    ) 
    , mcConfig_(cfg) { 
        mcConfig_.put("loopback", true); //always allow advertising to loopback, so other process
                                         //on the same machine get them
        mcConfig_.put("topicRegex", tcpcastAdTopic_.c_str());
        // mcConfig_.put("ifaceAddr", cfg.getExt<string>("ifaceAddr"));
        mcConfig_.put("outBufferSizePower2", 3u);
        // mcConfig_.put("mcastPort", cfg.getExt<uint16_t>("mcastPort"));
        // mcConfig_.put("mcastAddr", cfg.getExt<string>("mcastAddr"));
        // mcConfig_.put("ttl", cfg.getExt<int>("ttl"));

        mcSendTransport_.reset(new mcast::SendTransport(
            mcConfig_, sizeof(MessageWrap<TopicSource>)));
    }

    template <typename... Messages>
    void queue(Topic const& t, Messages&&... msgs) {
        auto n = sizeof...(msgs);
        auto it = buffer_.claim(n);
        queue(it, t, std::forward<Messages>(msgs)...);
        buffer_.commit(it, n);
    }

    template <typename... Messages>
    bool tryQueue(Topic const& t, Messages&&... msgs) {
        auto n = sizeof...(msgs);
        auto it = buffer_.tryClaim(n);
        if (it) {
            queue(it, t, std::forward<Messages>(msgs)...);
            buffer_.commit(it, n);
            return true;
        }
        return false;
    }

    template <typename Message, typename ... Args>
    void queueInPlace(Topic const& t, Args&&... args) {
        static_assert(!is_base_of<hasMemoryAttachment, Message>::value 
            || pattern::is_first_base_of<hasMemoryAttachment, Message>::value
            , "hasMemoryAttachment has to the first base for Message");
        auto s = buffer_.claim();
        char* addr = static_cast<char*>(*s);
        auto h = reinterpret_cast<TransportMessageHeader*>(addr);
        h->flag = calculateFlag<Message>();
        h->messagePayloadLen = sizeof(MessageWrap<Message>);
        auto tl = t.copyTo(addr + sizeof(TransportMessageHeader));
        h->topicLen = tl;
        if (hmbdc_likely(sizeof(Message) <= maxMessageSize_)) {
            new (addr + sizeof(TransportMessageHeader) + tl) MessageWrap<Message>(std::forward<Args>(args)...);
        } else {
            HMBDC_THROW(std::out_of_range
                , "maxMessageSize too small to hold a message when constructing SendTransportEngine");
        }
        buffer_.commit(s);
    }

    void queueBytes(Topic const& t, uint16_t tag, void const* bytes, size_t len) {
        auto s = buffer_.claim();
        char* addr = static_cast<char*>(*s);
        auto h = reinterpret_cast<TransportMessageHeader*>(addr);
        h->flag = 0;
        h->messagePayloadLen = sizeof(MessageWrap<JustBytes>) - 1 + len;
        auto tl = t.copyTo(addr + sizeof(TransportMessageHeader));
        h->topicLen = tl;
        if (hmbdc_likely(len <= maxMessageSize_)) {
            new (addr + sizeof(TransportMessageHeader) + tl) MessageWrap<JustBytes>(tag, bytes, len);
        } else {
            HMBDC_THROW(std::out_of_range
                , "maxMessageSize too small to hold a message when constructing SendTransportEngine");
        }
        buffer_.commit(s);
    }

    void stop() {
        buffer_.reset();
    };

    void initInThread() {
        mcSendTransport_->initInThread();
        Transport::initInThread();
    }

    bool match(Topic const& t) const {
        return boost::regex_match(t.c_str(), topicRegex_);
    }

private:
    std::string topic_;
    boost::regex topicRegex_;

protected:
    size_t maxMessageSize_;
    size_t minRecvToStart_;
    MonoLockFreeBuffer buffer_;

    unique_ptr<mcast::SendTransport> mcSendTransport_;
    Rater rater_;
    Config mcConfig_;

private:

    template <typename Message>
    static
    uint8_t calculateFlag() {
        if (is_base_of<hasMemoryAttachment, Message>::value) return hasMemoryAttachment::flag;
        return 0;
    }

    template<typename M, typename ... Messages>
    void queue(MonoLockFreeBuffer::iterator it, Topic const& t
        , M&& m, Messages&&... msgs) {
        using Message = typename std::remove_reference<M>::type;
        static_assert(!is_base_of<hasMemoryAttachment, Message>::value 
            || pattern::is_first_base_of<hasMemoryAttachment, Message>::value
            , "hasMemoryAttachment has to the first base for Message");

        auto s = *it;
        char* addr = static_cast<char*>(s);
        auto h = reinterpret_cast<TransportMessageHeader*>(addr);
        h->flag = calculateFlag<Message>();
        h->messagePayloadLen = sizeof(MessageWrap<Message>);
        auto tl = t.copyTo(addr + sizeof(TransportMessageHeader));
        h->topicLen = tl;
        if (hmbdc_likely(sizeof(Message) <= maxMessageSize_)) {
            new (addr + sizeof(TransportMessageHeader) + tl) MessageWrap<Message>(std::forward<M>(m));
        } else {
            HMBDC_THROW(std::out_of_range, "maxMessageSize too small to hold a message when constructing SendTransportEngine");
        }
        queue(++it, t, std::forward<Messages>(msgs)...);
    }

    void queue(MonoLockFreeBuffer::iterator it
        , Topic const& t) {}
};

struct SendTransportEngine
: SendTransport
, TimerManager
, ReoccuringTimer 
, single_thread_powered_client
, Client<SendTransportEngine> {  
    SendTransportEngine(Config const& cfg
        , size_t maxMessageSize)
    : SendTransport(cfg, maxMessageSize)
    , ReoccuringTimer(Duration::seconds(cfg.getExt<size_t>("topicAdvertisePeriodSeconds")))
    , mtu_(cfg.getExt<size_t>("mtu") - 20 - 20) //20 bytes ip header and 20 bytes tcp header
    , maxSendBatch_(config_.getExt<size_t>("maxSendBatch"))
    , waitForSlowReceivers_(cfg.getExt<bool>("waitForSlowReceivers"))
    , lastSeq_(0)
    , stopped_(false) {
        auto totalHead = sizeof(app::MessageHead) 
            + sizeof(Topic) + sizeof(TransportMessageHeader);
        if (maxMessageSize_ + totalHead > mtu_) {
            HMBDC_THROW(std::out_of_range, "maxMessageSize needs <= " << mtu_ - totalHead);
        }
        toSend_.reserve(maxSendBatch_);
        // inititialize begin_ to first item
        MonoLockFreeBuffer::iterator end;
        buffer_.peek(begin_, end, 1u); 
        buffer_.wasteAfterPeek(begin_, 0, true);
    }
    using SendTransport::SendTransport;

    char const* hmbdcName() const { 
        return this->hmbdcName_.c_str();
    }

    std::tuple<char const*, int> schedSpec() const {
        return std::make_tuple(this->schedPolicy_.c_str(), this->schedPriority_);
    }

    /*virtual*/ 
    void invokedCb(uint16_t threadSerialNumber) HMBDC_RESTRICT override  {
        //only START to send when enough sessions are ready to receive or
        //buffer is full
        if (hmbdc_likely(!minRecvToStart_ 
            || asyncServer_->readySessionCount() >= minRecvToStart_)) {
            minRecvToStart_ = 0; //now started no going back
            runAsync();
        }
        mcSendTransport_->runOnce();

        if (hmbdc_unlikely(pIos_->stopped())) pIos_->reset();
        boost::system::error_code ec;
        pIos_->poll(ec);
        if (hmbdc_unlikely(ec)) HMBDC_LOG_C("io_service error=", ec);
    }

    /*virtual*/ 
    bool droppedCb() override {
        stop();
        stopped_ = true;
        return true;
    };

    /*virtual*/ 
    void messageDispatchingStartedCb(uint16_t threadSerialNumber) override {
        this->initInThread();
        asyncServer_.reset(new AsyncSendServer(config_, *pIos_, buffer_.capacity()));
        setCallback(
            [this](TimerManager& tm, SysTime const& now) {
                mcSendTransport_->queue(
                    tcpcastAdTopic_, asyncServer_->advertisingMessage());
                if (!waitForSlowReceivers_) cullSlow();
            }
        );

        schedule(SysTime::now(), *this);
    }

    /**
     * @brief check how many recipient sessions are still active
     * @return numeric_limits<size_t>::max() if the sending hasn't started due to 
     * minRecvToStart has not been met yet
     */
    size_t sessionsRemainingActive() const {
        if (stopped_) return 0;
        if (asyncServer_ && !minRecvToStart_) {
            return asyncServer_->readySessionCount();
        }
        return numeric_limits<size_t>::max();
    }
private:
    void cullSlow() {
        auto seq = buffer_.readSeq();
        if (seq == lastSeq_ && buffer_.isFull()) {
            asyncServer_->killSlowestSession();
        }
        lastSeq_ = seq;
    }

    void runAsync() HMBDC_RESTRICT {
        MonoLockFreeBuffer::iterator begin, end;
        buffer_.peek(begin, end, maxSendBatch_);
        auto it = begin_;
        pair<char const*, char const*> currentTopic;
        currentTopic.first = nullptr;
        size_t currentTopicLen = 0;
        size_t toSendByteSize = 0;

        while (it != end) {
            void* ptr = *it;
            auto item = static_cast<TransportMessageHeader*>(ptr);

            if (hmbdc_unlikely(!rater_.check(item->wireSize()))) break;
            if (hmbdc_unlikely(!item->flag 
                && toSendByteSize + item->wireSize() > mtu_)) break;
            it++;
            if (hmbdc_unlikely(!currentTopic.first)) {
                currentTopic = item->topic();
                currentTopicLen = item->topicLen;
                toSendByteSize = item->wireSize();
                toSend_.emplace_back(ptr, item->wireSize());
            } else if (item->topicLen == currentTopicLen 
                && strncmp(currentTopic.first, item->topic().first, currentTopicLen)
                    == 0) {          
                toSendByteSize += item->wireSize();
                toSend_.emplace_back(ptr, item->wireSize());
            } else {
                asyncServer_->queue(currentTopic
                    , move(toSend_)
                    , toSendByteSize
                    , it);
                toSend_.clear();

                currentTopic = item->topic();
                currentTopicLen = item->topicLen;
                toSendByteSize = item->wireSize();
                toSend_.emplace_back(ptr, item->wireSize());
            }

            if (hmbdc_unlikely(item->flag == hasMemoryAttachment::flag)) {
                auto& a = item->wrapped<hasMemoryAttachment>();
                toSend_.emplace_back(a.attachment, a.len);
                toSendByteSize += a.len;
            }
            rater_.commit();
        }

        if (toSend_.size()) {
            asyncServer_->queue(currentTopic
                , move(toSend_)
                , toSendByteSize
                , it);
            toSend_.clear();
        }
        begin_ = it;
        //use it here which is the maximum
        auto newStart = asyncServer_->run(begin, it);
        if (begin != end) { //only need to do this when there r things read from buffer_
            for (auto it = begin; it != newStart; ++it) {
                void* ptr = *it;
                auto item = static_cast<TransportMessageHeader*>(ptr);
                if (hmbdc_unlikely(item->flag == hasMemoryAttachment::flag)) {
                    auto& a = item->wrapped<hasMemoryAttachment>();
                    if (a.afterConsumedCleanupFunc) a.afterConsumedCleanupFunc(&a);
                }
            }
            buffer_.wasteAfterPeek(begin, newStart - begin, true);
        }
    }

    size_t mtu_;
    size_t maxSendBatch_;
    bool waitForSlowReceivers_;
    MonoLockFreeBuffer::iterator begin_;
    vector<const_buffer> toSend_;
    // SendServer *server_;
    // unique_ptr<SyncSendServer> syncServer_;
    unique_ptr<AsyncSendServer> asyncServer_;
    size_t lastSeq_;
    bool stopped_;
};
} //send_detail

using SendTransport = send_detail::SendTransport;
using SendTransportEngine = send_detail::SendTransportEngine;
}}}
